/* KESKIDI MODULE by Artix 
	Version 1.0.0
	Based on the original Keskidi TCL Script by Artix. (hey, that's me again!)
*/

/*
 That module is a replacement for the v1 Keskidi TCL script,
 and designed to work with the v2 one.
 
 The script returns a random line said by a selected
 person based on the eggdrop logs.
 However, due to the way TCL works (extensive debugguing, 
 command interpretation, and so on) the search time would 
 rise exponentially, reaching half a second on a single search
 over only a week on very active channels.
 
 This module is designed to work with the version 2 script.
 The script will search for the files to open using the configuration
 file's variables and stuff, then ask the module to search the line for him.
*/

// Eggdrop definition stuff
#define MAKING_KESKIDI
#define MODULE_NAME "keskidi"
#define VER1 1
#define VER2 0

// Basic includes
// There might be unused ones though...
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
// Header include
#include "keskidi.h"
// Eggdrop include
#include "../module.h"

// I don't have a clue what the hell that thing is.
// But i guess it's important, as it's present in all other modules.
#undef global
static Function *global = NULL;

// Maximum charracters on a line from the logs
#define MAXLINECHARS 2000
// Number of lines to allocate at a time
// more would inecrase memory usage, and less
// would inecrase cpu usage/execution time.
#define ALLOCBLOCK 100
// Max nick chars to be alloc'd
#define MAXNICKCHARS 40

// The TCL binding
static tcl_cmds mytcls[] =
{
  {"keskidi_search", tcl_search},
  {0, 0}
};

// Let's initialize the command call counter
// That thing is then used to generate
// the random generator's seed.
static int call = 0;

// Define a string variable for the nick
// The Compiler didn't like me passing a local pointer 
// to another function, so i prefered using an external variable.
static char checknick[MAXNICKCHARS];

// Definition of the functions to use
// by the module system of the eggdrop
// (at load, unload, using .module, and so on)
static Function keskidi_table[] =
{
  (Function) keskidi_start,
  (Function) keskidi_close,
  (Function) keskidi_expmem,
  (Function) keskidi_report,
};


static int keskidi_expmem()
{
	/* All variables are alloc'd and free'd within the
	 same function, so there should be no measurable
	 memory usage... i guess. 
	Of course we could measure the size of the 
	few integers stored here and here, but that'd be like
	lieing, as the module obviously uses more memory. */
	return 0;
}

// Function called when unloading the module
static char *keskidi_close()
{
	// Remove the TCL Binding
	rem_tcl_commands(mytcls);
	// Unload the module by itself
	module_undepend(MODULE_NAME);
	return NULL;
}

// Function called when loading the module
char *keskidi_start(Function *global_funcs)
{
	// Register the module in the eggdrop stuff
	global = global_funcs;
	module_register(MODULE_NAME, keskidi_table, VER1, VER2);
	// Bind the TCL Commands
	add_tcl_commands(mytcls);

	// There, we can show a pretty message.
	putlog(LOG_MISC, "*", "Keskidi Module v%i.%i by Artix loaded successfully.", VER1, VER2); 
	return NULL;
}

// A report on the module's status using .module on the PL
static void keskidi_report(int idx, int details)
{
	if (details) {
          dprintf(idx, "     Keskidi script's module");
	}
}

// The main module's command
// It's called by the TCL Interpreters
// Arguments: Nickname of the people we are searching,
//	      N More arguments for every file to search in
static int tcl_search STDVAR 
{
	// Initialize integers, pointers, etc.
	int i = 0;
	int j = 0;
	FILE *filepointer = NULL;
	char **filelist = NULL; 
	char readline[MAXLINECHARS];
	struct POSINFO *poslist = NULL; 
	int randentry = 0;
	int currentpos = 0;
	int seed = 0;
	// Add one to the call counter
	call++;
	// Just so that the randomly calculated stuff
	// don't get over int max value or something like that
	if(call > 10000) call=0;

	// Let's verify the syntax first
	// NOTE: Maybe a proper use of BADARGS is better than checking argc
	if(argc < 3) BADARGS(3, 3, " nickname file1 .... ?fileN?");
	
	// Allocate the filelist based on the number
	// of arguments (therefore, files) given by TCL
	if((filelist = nmalloc(sizeof(char*)*(argc-1))) == NULL) {
		putlog(LOG_MISC, "=== Keskidi Module Warning: Unable to allocate the file path list"); 
		return TCL_OK;
	}

	// Loop for every file
	for(i=2; i < argc; i++) 
	{
		// First we need to get an entry in the filelist for the file
		if((filelist[i-2] = nmalloc(sizeof(char)*(strlen(argv[i])+1))) == NULL) {
			putlog(LOG_MISC, "=== Keskidi Module Warning: Unable to allocate file path entry %i", i-2);
			freefile(filelist, argc);
			if(poslist != NULL) nfree(poslist);
			return TCL_ERROR;
		}

		// Then we can strcpy the path into it's filelist entry
		strncpy(filelist[i-2], argv[i], strlen(argv[i])+1);
	
		// We then try to open the file (so that the entries in filelist and i are same)
		if ((filepointer = fopen(argv[i], "r")) == NULL) continue;
		
		currentpos = 0;
		while(fgets(readline, MAXLINECHARS, filepointer) != NULL) 
		{
			// First we check the line using another proc.
			// Note that it do store the nickname of the one who said that line in
			// the external variable checknick.
			checkline(readline);
			if((strcmp(checknick,argv[1]) == 0) || ((strlen(checknick) > 0) && (argv[1][0] == '*'))) 
			{
				// Once we reach the end of our alloc'd number of
				// entries, we got to realloc the array to get more aviable.
				if((j%ALLOCBLOCK) == 0)
				{
					if((poslist = realloc(poslist, sizeof(struct POSINFO)*(j+ALLOCBLOCK))) == NULL) {
						putlog(LOG_MISC, "=== Keskidi Module Warning: Unable to reallocate the position list");
						// As usual, if there is a memory allocation problem, we clean up the stuff and return.
						freefile(filelist, argc);
						fclose(filepointer);
						nfree(poslist); 
						return TCL_OK; 
					}   
				}

				// Add the current line's position into the list.
				poslist[j].fileid = i-2;
				poslist[j].pos = currentpos;
				j++;
			}
		currentpos = ftell(filepointer);
		}
		
		// Finally, we exit the file. 
		fclose(filepointer);
	}

	// In case we haven't found anything, we can just stop here.
	if(j==0) {
		freefile(filelist, argc);
		return TCL_OK;
	}


	// Calculate the Random Generator's seed.
	// It's not really good, but i couldn't  get anything better and it seems to work
	// Please send me any suggestions.
	seed = (int)(fabs (sin ((double)call*31.0))*4267.0);
	srand(seed);
	
	// Get a random entry in the position list
	randentry = (int)((double)rand()/(double)RAND_MAX*(double)j);

	// Reopen the corresponding file
	if((filepointer = fopen(filelist[poslist[randentry].fileid], "r")) == NULL) {
		putlog(LOG_MISC, "*", "=== Keskidi Module Warning: Unable to re-open that file for an unknown reason: %s", filelist[poslist[randentry].fileid]);
		freefile(filelist, argc);
		nfree(poslist); 
		return TCL_OK;
	}	
 
	// Get to the corresponding location
	fseek(filepointer, (poslist[randentry].pos), SEEK_SET); 
	
	// Read the line and return it
	if((fgets(readline, MAXLINECHARS, filepointer)) != NULL) 
	{
		// We pass the result to the TCL Interpreter as a string.
		Tcl_SetObjResult(irp, Tcl_NewStringObj(readline, -1));
	}

	// Cleanup time.
	fclose(filepointer); 
	freefile(filelist, argc);
	nfree(poslist);
	return TCL_OK;
}

// Free the whole filelist, meaning both sub-strings and the pointer
static void freefile(char **filelist, int args) {
	int i = 0; 
	args = args - 2; // There's the TCL Command and the nickname so we have to remove 2 to argc to get the number of files.
	for (i=0; i < args; i++) nfree(filelist[i]);
	nfree(filelist);
}

static char *checkline(char *readline) 
{
	int linelen = strlen(readline);
	int ci = 0;
	int cj = 0;

	checknick[0] = '\0';

	if(readline[0] == '[') {
		for(ci=1; ci<linelen; ci++) {
			if(readline[ci] == ']') {
				ci = ci+2; 
				break;
			}
		}
	}

	if(readline[ci] == '<') {
		ci++;
		while(ci<linelen) {
			if(readline[ci] == '>') {
				checknick[cj] = '\0';
				break;
			}
		checknick[cj] = readline[ci];
		cj++;
		ci++;
		}
	}

	if(cj == 0) return; 
}	
